package org.apache.bcel.classfile;

import java.util.stream.Stream;
import java.util.Iterator;
import java.io.DataOutputStream;
import java.util.Arrays;
import java.io.IOException;
import java.io.DataInput;

public final class LineNumberTable extends Attribute implements Iterable<LineNumber>
{
    private static final int MAX_LINE_LENGTH = 72;
    private LineNumber[] lineNumberTable;
    
    LineNumberTable(final int nameIndex, final int length, final DataInput input, final ConstantPool constantPool) throws IOException {
        this(nameIndex, length, (LineNumber[])null, constantPool);
        final int lineNumberTableLength = input.readUnsignedShort();
        this.lineNumberTable = new LineNumber[lineNumberTableLength];
        for (int i = 0; i < lineNumberTableLength; ++i) {
            this.lineNumberTable[i] = new LineNumber(input);
        }
    }
    
    public LineNumberTable(final int nameIndex, final int length, final LineNumber[] lineNumberTable, final ConstantPool constantPool) {
        super((byte)4, nameIndex, length, constantPool);
        this.lineNumberTable = lineNumberTable;
    }
    
    public LineNumberTable(final LineNumberTable c) {
        this(c.getNameIndex(), c.getLength(), c.getLineNumberTable(), c.getConstantPool());
    }
    
    @Override
    public void accept(final Visitor v) {
        v.visitLineNumberTable(this);
    }
    
    @Override
    public Attribute copy(final ConstantPool constantPool) {
        final LineNumberTable c = (LineNumberTable)this.clone();
        Arrays.setAll(c.lineNumberTable = new LineNumber[this.lineNumberTable.length], i -> this.lineNumberTable[i].copy());
        c.setConstantPool(constantPool);
        return c;
    }
    
    @Override
    public void dump(final DataOutputStream file) throws IOException {
        super.dump(file);
        file.writeShort(this.lineNumberTable.length);
        for (final LineNumber lineNumber : this.lineNumberTable) {
            lineNumber.dump(file);
        }
    }
    
    public LineNumber[] getLineNumberTable() {
        return this.lineNumberTable;
    }
    
    public int getSourceLine(final int pos) {
        int l = 0;
        int r = this.lineNumberTable.length - 1;
        if (r < 0) {
            return -1;
        }
        int minIndex = -1;
        int min = -1;
        do {
            final int i = l + r >>> 1;
            final int j = this.lineNumberTable[i].getStartPC();
            if (j == pos) {
                return this.lineNumberTable[i].getLineNumber();
            }
            if (pos < j) {
                r = i - 1;
            }
            else {
                l = i + 1;
            }
            if (j >= pos || j <= min) {
                continue;
            }
            min = j;
            minIndex = i;
        } while (l <= r);
        if (minIndex < 0) {
            return -1;
        }
        return this.lineNumberTable[minIndex].getLineNumber();
    }
    
    public int getTableLength() {
        return (this.lineNumberTable == null) ? 0 : this.lineNumberTable.length;
    }
    
    @Override
    public Iterator<LineNumber> iterator() {
        return Stream.of(this.lineNumberTable).iterator();
    }
    
    public void setLineNumberTable(final LineNumber[] lineNumberTable) {
        this.lineNumberTable = lineNumberTable;
    }
    
    @Override
    public String toString() {
        final StringBuilder buf = new StringBuilder();
        final StringBuilder line = new StringBuilder();
        final String newLine = System.getProperty("line.separator", "\n");
        for (int i = 0; i < this.lineNumberTable.length; ++i) {
            line.append(this.lineNumberTable[i].toString());
            if (i < this.lineNumberTable.length - 1) {
                line.append(", ");
            }
            if (line.length() > 72 && i < this.lineNumberTable.length - 1) {
                line.append(newLine);
                buf.append((CharSequence)line);
                line.setLength();
            }
        }
        buf.append((CharSequence)line);
        return buf.toString();
    }
}
